# coding: utf-8
# These functions are generated by ChatGPT (https://chat.openai.com/chat), Dec 15 [2022] Version.
# According to ChatGPT, the code is licensed under MIT license.

from __future__ import unicode_literals
from gqlspection import log
from builtins import str


def minimize_query(query):
    # Set up a buffer for the minimized query
    minimized = ""

    # Set up a flag to track whether we are inside a comment
    in_comment = False

    # Set up a flag to track whether we are inside a string
    in_single_string = False
    in_double_string = False
    in_string = False

    # Set up a flag to track whether we are inside a whitespace
    in_space = False

    # Set up a flag to track whether the current character is an escaped quote
    escaped = False

    # Iterate over each character in the query
    for i, c in enumerate(query.strip()):
        # If we see the start of a comment and we are not inside a string,
        # set the in_comment flag
        if c == "#" and not in_string:
            in_comment = True

        # Preserve strings verbatim
        if in_string:
            minimized += c
        # Skip comments entirely
        elif not in_comment:
            # Collapse all whitespace into one space character
            if in_space:
                if not c.isspace():
                    minimized += " " + c
            elif not c.isspace():
                minimized += c

        # If we are inside a whitespace, set the in_space flag
        in_space = c.isspace()

        # If we see the end of a comment and we are not inside a string,
        # clear the in_comment flag
        if not in_string and (c == "\n"):
            in_comment = False

        # Toggle the in_*_string flags if we see quote character, unless it's escaped
        if not in_comment and not escaped:
            in_double_string = not in_double_string if c == '"' else in_double_string
            in_single_string = not in_single_string if c == "'" else in_single_string
        in_string = in_single_string or in_double_string

        # Set the escaped flag, if the current character is a backslash
        escaped = (c == "\\")

    # Return the minimized query
    return minimized


def format_comment(string, max_length=60):
    # Split the string into lines
    lines = string.split('\n')

    # Initialize a list to store the formatted lines
    formatted_lines = []

    # Iterate through the lines
    for line in lines:
        # Split the line into words
        words = line.split()

        # Initialize a variable to keep track of the current line
        current_line = ''

        # Iterate through the words
        for word in words:
            # If the current line plus the next word would be too long,
            # append the current line to the list of formatted lines and start a new line
            if len(current_line) + len(word) > max_length:
                formatted_lines.append(current_line)
                current_line = ''

            # If the current line is empty, add the word to the line
            # Otherwise, add a space and the word to the line
            if current_line == '':
                current_line = word
            else:
                current_line += ' ' + word

        # Add the remaining line to the list of formatted lines
        formatted_lines.append(current_line)

    # Return the formatted comment
    return '\n'.join(['# ' + line for line in formatted_lines])


def safe_get_list(dictionary, name):
    """Safely extract list from the dictionary, even if the key does not exist or type is wrong."""
    res = dictionary.get(name, [])

    if type(res) is list:
        return res
    else:
        return []


def pad_string(string, n=4):
    """Pad a multiline string with n spaces."""
    if not n or not string:
        return string

    if isinstance(string, str):
        ends_with_newline = string[-1] == '\n'
        return '\n'.join((
            (' ' * n + line) for line in string.splitlines()
        )) + ('\n' if ends_with_newline else '')
    else:
        log.debug("Asked to pad the following non-string with %s spaces: %s", n, string)
        raise Exception("Expected a string to pad, received %s", type(string))



# coding: utf-8
from __future__ import unicode_literals


__documented_introspection_query = """
query IntrospectionQuery {
    __schema {
        # Typically query is called "Query" and mutation "Mutation", but those can be redefined.
        # For some reason, spec does not force queryType[name] to be String!, but I don't think it can be null.
        queryType {
            name
            }
        }
        # 'mutationType' can be null if there are no mutations.
        mutationType {
            name
        }
        # TODO: We're not parsing subscriptions and directives at all right now
        # subscriptionType { name }
        # directives { name }
        types {
            name
            # 'kind' is enum with values: SCALAR, OBJECT, INTERFACE, UNION, ENUM, INPUT_OBJECT, LIST, NON_NULL
            kind
            description
            # The following are only present for OBJECT and INTERFACE, otherwise null:
            fields(includeDeprecated: true) {
                name
                description
                args(includeDeprecated: true) {
                    ... InputValue
                }
                type {
                    ... TypeRef
                }
                isDeprecated
                deprecationReason
            }
            interfaces {
                ... TypeRef
            }
            # The following is only non-null for INTERFACE and UNION:
            possibleTypes {
                ... TypeRef
            }
            # The following is only non-null for ENUM:
            enumValues(includeDeprecated: true) {
                name
                description
                isDeprecated
                deprecationReason
            }
            # The following is only non-null for INPUT_OBJECT:
            inputFields(includeDeprecated: true) {
                ... InputValue
            }
            # The following is only non-null for LIST and NON_NULL:
            ofType {
                ... TypeRef
            }
            # Only (optionally) non-null for custom scalars:
            specifiedByURL
        }
    }
}

fragment InputValue on __InputValue {
  name
  description
  type { ...TypeRef }
  defaultValue
}

# TODO: generate this query dynamically and make the depth adjustable
fragment TypeRef on __Type {
  kind
  name
  ofType {
    kind
    name
    ofType {
      kind
      name
      ofType {
        kind
        name
        ofType {
          kind
          name
          ofType {
            kind
            name
            ofType {
              kind
              name
              ofType {
                kind
                name
              }
            }
          }
        }
      }
    }
  }
}
"""


# TODO: 'depth' is ignored right now, fix it
def get_introspection_query(depth=4, minimize=True):
    """Construct the introspection query and optionally minimize it."""
    from gqlspection.utils import minimize_query
    if minimize:
        return minimize_query(__documented_introspection_query)
    else:
        return __documented_introspection_query



# coding: utf-8
from __future__ import unicode_literals
from builtins import object, str
from gqlspection import log
import gqlspection


class GQLSchema(object):
    types          = None
    query          = None
    mutation       = None

    def __init__(self, url=None, extra_headers=None, json=None, logger=None):
        if logger:
            log.logger = logger

        if json:
            if isinstance(json, str):
                json = self._str_to_json(json)

            if 'data' in json and '__schema' in json['data']:
                original_schema = json['data']['__schema']
            elif '__schema' in json:
                original_schema = json['__schema']
            elif 'types' in json:
                original_schema = json
            else:
                raise Exception("GQLSchema: Couldn't parse JSON schema.")
        elif url:
            introspection_result = self.send_request(url, extra_headers)
            original_schema = introspection_result['data']['__schema']
        else:
            raise Exception("GQLSchema: Provide either JSON or URL.")

        self.types = gqlspection.GQLTypes(self, original_schema)

        self.query    = self._extract_query_type(original_schema)
        self.mutation = self._extract_mutation_type(original_schema)

    @staticmethod
    def _str_to_json(data):
        import json
        return json.loads(data)

    def _extract_query_type(self, schema):
        """Get the query type name (typically 'Query').

        According to GraphQL specs, this should always be present. We could easily fall back to 'Query' though, file
        a bug report if there is a real-world need for that.

            __schema {
                queryType {
                    name
                }
            }
        """
        if 'queryType' in schema:
            name = schema['queryType'].get('name', False)
            if name:
                return gqlspection.GQLTypeProxy(name, self)
            else:
                raise Exception("GQLSchema: Invalid schema - queryType is null. File a bug if this happens in real "
                                "world.")
        else:
            raise Exception("GQLSchema: Invalid schema - no queryType (absence of queryType is not allowed according "
                            "to GraphQL spec, file a bug report if this happens in real world).")

    def _extract_mutation_type(self, schema):
        """Get the mutation type and return None if it has not been defined.

            __schema {
                mutationType {
                    name
                }
            }
        """
        if 'mutationType' in schema:
            name = schema['mutationType'].get('name', None)
            if name:
                return gqlspection.GQLTypeProxy(name, self)
            else:
                raise None
        else:
            return None

    @staticmethod
    def send_request(url, extra_headers=None, minimize=True):
        import requests
        from gqlspection.introspection_query import get_introspection_query

        headers = {'Content-Type': 'application/json'}
        if extra_headers:
            headers.update(extra_headers)

        result = requests.post(url, json={'query': get_introspection_query(minimize=minimize)}, headers=headers).json()
        if 'errors' in result:
            raise Exception([error['message'] for error in result['errors']])

        return result

    def generate_query(self, name):
        if isinstance(name, str):
            field = self.query.fields[name]
        else:
            field = name
        return gqlspection.GQLQuery(self.query, 'query', fields=[field])

    def generate_mutation(self, name):
        if isinstance(name, str):
            field = self.query.fields[name]
        else:
            field = name
        return gqlspection.GQLQuery(self.mutation, 'mutation', fields=[field])



# coding: utf-8
from __future__ import unicode_literals
from setuptools import setup, find_packages
import os

pwd = os.path.abspath(os.path.dirname(__file__))
readme_path = os.path.join(pwd, "README.md")

with open(readme_path) as f:
    readme = f.read()


def is_jython():
    from platform import python_implementation
    return python_implementation() == 'Jython'


if is_jython():
    # Create _version.py ahead of type by running:
    #   python3 -c "import setuptools_scm; setuptools_scm.get_version(write_to='_version.py')"
    from _version import version

setup(
    name="gqlspection",
    # If Jython, import version from version.py file (should be created ahead of time)
    version=version if is_jython() else None,
    # Otherwise, load version from git tags
    setup_requires=[
        "setuptools_scm"
    ] if not is_jython() else [],
    use_scm_version=not is_jython(),
    description="GraphQL Introspection parsing and query generation",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/doyensec/gqlspection",
    author="Andrew Konstantinov",
    author_email="andrew@doyensec.com",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "Environment :: Console",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: Jython",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Software Development :: Code Generators",
        "Topic :: Security"
    ],
    keywords="graphql, introspection",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=2.7",
    install_requires=[
        "pathlib2; python_version == '2.7'",
        "future;   python_version == '2.7'"
    ],
    extras_require={
        "cli": ["click", "requests"]
    },
    entry_points={
        "console_scripts": [
            "gqlspection=gqlspection:cli"
        ]
    },
    project_urls={
        "Source": "https://github.com/doyensec/gqlspection",
        "Bug Reports": "https://github.com/doyensec/gqlspection/issues",
        "Doyensec Research": "https://www.doyensec.com/research.html"
    }
)



# coding: utf-8
# noinspection PyUnresolvedReferences
from __future__ import unicode_literals

try:
    from pathlib import Path
except ImportError:
    from pathlib2 import Path

from gqlspection import GQLSchema


def pytest_generate_tests(metafunc):
    if 'name' in metafunc.fixturenames:
        pwd = Path(__file__).parent
        names = (x.stem for x in pwd.glob('data/*.json'))
        metafunc.parametrize("name", names)
        metafunc.parametrize("mode", ('queries', 'mutations'))


def test_start(name, mode):
    pwd = Path(__file__).parent / 'data'

    # Load expected results (if the file is present)
    path = pwd / "{}.{}.txt".format(name, mode)
    if not path.is_file():
        return
    expected_results = path.read_text()

    # Load the GraphQL schema
    path = pwd / "{}.json".format(name)
    schema = GQLSchema(json=path.read_text())

    # generate a string representing queries / mutations
    if mode == 'queries':
        result = '\n'.join(schema.generate_query(field).str() for field in schema.query.fields)
    else:
        result = '\n'.join(schema.generate_mutation(field).str() for field in schema.mutation.fields)

    print(type(result), type(expected_results))

    assert result.strip() == expected_results.strip()
